/* tones.c
 * generate and play tones for given period....
 * Jim Jackson     Linux Version
 */

/* compilation
 *
 *   cc -o tones tones.c generator.c misc.c -lm
 */

/*
 * Copyright (C) 1997 Jim Jackson                    jj@scs.leeds.ac.uk
 *                    School of Computer Studies,
 *                    The University of Leeds,
 *                    Leeds, LS2 9JT, UK
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program - see the file COPYING; if not, write to 
 *  the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 *  MA 02139, USA.
 */

/* History...................
 * 
 * 13Feb97  Based on sgen, using generator.c - only mono version.
 * 16Feb97  altered it so you can specify several freqs to be played together
 *          by using punctuation of some sort to seperate freqs 
 *          e.g. 440,880   would play two freqs of 440Hz and 880Hz together
 * 
 */

#include <stdio.h>
#include <strings.h>
#include <errno.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <math.h>
#include "fmtheaders.h"

#define VERSION "tones  Ver. 1.0 (Mar 97)   tone generator"

#define ABUF_SIZE 4096
#define DAC_FILE "/dev/dsp"
#define SAMPLERATE 22050
#define EFAIL -1

#define MAXPRM 32
#define MAXTONES 64
#define chkword(a,b) ((n>=a)&&(strncmp(p,b,n)==0))

int aflg,fflg,vflg,dflg,wfmt;
int loopflg;

unsigned int stereo;             /* stereo or mono mode      */
unsigned int samplerate;         /* Samples/sec              */
unsigned int ratio;              /* used in pulse, sweep etc */
unsigned int afmt;               /* format for DSP  */
unsigned int abuf_size;          /* audio buffer size */
unsigned int tbuf_size;          /* tone buffer size */
unsigned int DSPbytecount;       /* number of bytes sent to DSP  */
int frag_size;                   /* size of a DSP buffer  */

char *waveform;                  /* Waveform to generate for tones    */
unsigned int Tms;                /* period in millisecs for each tone */
unsigned int Ntone;              /* number of samples for Tms         */
unsigned int tfreq[MAXTONES];    /* frequencies to generate           */
unsigned char *tones[MAXTONES];  /* array of buffers containing tone samples */
int Nt;                          /* Number of tones in the tones[] array */
unsigned char *silence;          /* silence buffer   */

char *sys,*outfile;
unsigned char *maketone();

help(e)
int e;
{  fputs(VERSION,stderr);
   fputs("\nGenerates and plays for T millisecs, tones of given frequencies.\n",stderr);
   fputs("An optional waveform can be specified - sine, cosine, square, pulse,\n",stderr);
   fputs("triangle, sawtooth.\n",stderr);
   fputs("Several freqs can be played together by specifying a list of freqs\n",stderr);
   fputs("separated by any punctuation character, e.g. 150,200,300 or 100+200+300\n",stderr);
   fputs("\nUsage: tones [-s samplerate] [-8/-16|-b 8/16] [waveform] T freq(s)\n",stderr);
   fputs("Defaults: sine wave, output to /dev/dsp, 22050 samples/sec, mono, 16 bit\n",stderr);
   fputs("          samples if possible, else 8 bit. \n\n",stderr);
   fputs("flags: -f,-a         force overwrite/append of/to file\n",stderr);
   fputs("       -o file       write digital sample to file ('-' is stdout)\n",stderr);
   fputs("       -w file       as '-o' but written as a wave file\n",stderr);
   fputs("       -s samples    generate with samplerate of samples/sec\n",stderr);
   fputs("       -loop         repeat play the tones ad nauseam.\n",stderr);
   fputs("       -v            be verbose.\n",stderr);
   fputs(" -8/-16 or -b 8|16   force 8 bit or 16 bit mode.\n",stderr);
   return(e);
}

/* main
 *
 */
 
main(argc,argv)
int argc;
char **argv;
{
   unsigned int i,j,k,l,m,n,N;
   FILE *f;
   char *p,*wf,*fnm,fname[130],*omode;
   int fd,st;
   unsigned int t;
   
   if ((p=strrchr(sys=*argv++,'/'))!=NULL) sys=++p;
   argc--;
   
   waveform="sine";
   outfile=NULL;
   samplerate=SAMPLERATE; afmt=AFMT_QUERY;
   loopflg=stereo=wfmt=0;                 /* force mono mode */
   dflg=vflg=aflg=fflg=0;
   
   while (argc && **argv=='-') {          /* all flags and options must come */
      n=strlen(p=(*argv++)+1); argc--;    /* before paramters                */
      if (chkword(1,"samplerate")) {
	 if (argc && isdigit(**argv)) { samplerate=atoi(*argv++); argc--; }
      }
      else if (chkword(1,"output")) {     /* specify output file       */
	 if (argc) { outfile=*argv++; argc--; } /* output file name          */
      }
      else if (chkword(1,"loop")) {       /*  specify loop option   */
         loopflg=1;
      }
      else if (chkword(1,"wave")) {     /* e.g. write WAVE format file  */
	 if (argc) { outfile=*argv++; argc--; wfmt=1; } /* output file name  */
      }
      else if (chkword(2,"16")) { afmt=AFMT_S16_LE; }
      else if (chkword(1,"bits")) {
	 i=0;
	 if (argc) { 
	    i=atoi(*argv++); argc--;
	 }
	 if (i==8) afmt=AFMT_U8;
	 else if (i==16) afmt=AFMT_S16_LE;
	 else exit(err_rpt(EINVAL,"must be '-b 8' or '-b 16'."));
      }
      else {                              /* check for single char. flags    */
	 for (; *p; p++) {
	    if (*p=='h') exit(help(EFAIL));
	    else if (*p=='8') afmt=AFMT_U8;
	    else if (*p=='a') aflg=1;
	    else if (*p=='d') dflg=1;
	    else if (*p=='l') loopflg=1;
	    else if (*p=='v') vflg=1;
	    else {
	       *fname='-'; *(fname+1)=*p; *(fname+2)=0;
	       exit(help(err_rpt(EINVAL,fname)));
	    }
	 }
      }
   }
   
   if (argc && !isdigit(**argv)) {
      waveform=*argv++; argc--;
   }
   
   if (argc<2) exit(help(err_rpt(EINVAL,"Wrong number of parameters")));
   Tms=atoi(*argv++); argc--;
   if (Tms==0) exit(err_rpt(EINVAL,"Millisecs should be > 0"));

   if (vflg) puts(VERSION);

   if (outfile==NULL) {           /* if no outfile then write direct to DAC */
                                  /* if no format specified then try 16 bit */
      i=(afmt==AFMT_QUERY)?AFMT_S16_LE:afmt ; 
      if ((fd=DACopen(fnm=DAC_FILE,"w",&samplerate,&i,&stereo))<0) {
	 if (afmt==AFMT_QUERY) {  /* if no format specified try for 8 bit.. */
	    i=AFMT_U8;
	    fd=DACopen(fnm=DAC_FILE,"w",&samplerate,&i,&stereo);
	 }
	 if (fd<0) exit(err_rpt(errno,"Opening DSP for output."));
      }
      afmt=i;
      if (vflg) {
	 fputs(fnm,stdout); fputs(": DAC Opened for output\n",stdout);
	 fputd(samplerate,stdout); fputs(" samples/sec, ",stdout);
	 if (i==AFMT_U8) { p="unsigned 8 bit samples\n"; }
	 else if (i==AFMT_S16_LE) { p="signed 16 bit samples, little endian\n"; }
	 else { p="Unknown audio format, "; }
	 fputs(p,stdout);
         fputd(frag_size,stdout); fputs(" bytes per DSP buffer.\n",stdout);
      }
      if (i!=afmt || stereo!=0) {
	 if (i!=afmt) 
	   err_rpt(EINVAL,"Sound card doesn't support requsted format.");
	 if (stereo!=0)
	   err_rpt(EINVAL,"Card doesn't do MONO - nah, don't believe it");
	 exit(EINVAL);
      }
   } else { 
      if (aflg && wfmt) {
         exit(err_rpt(EINVAL,"Cannot write append to a WAVE file."));
      }
      afmt=(afmt==AFMT_QUERY)?AFMT_S16_LE:afmt ; /* set 16 bit mode unless
						  * some format is forced */
      omode=(aflg)?"a":"w";
      if ((f=fopen(fnm=outfile,"r"))!=NULL) {
	 fclose(f);
	 if ( (!aflg) && (!fflg) ) exit(err_rpt(EEXIST,fnm));
      }
      if ((f=fopen(fnm,omode))==NULL) { exit(err_rpt(errno,fnm)); }
      if (vflg) { 
	 fputs(fnm,stdout);
	 if (*omode=='a') fputs(": Opened in append mode.\n",stdout);
	 else fputs(": Opened clean for writing.\n",stdout);
      }
      fd=fileno(f);
   }

   tbuf_size=samplerate;                  /* buf sizes if 8 bit samples */
   if (afmt==AFMT_S16_LE) tbuf_size<<=1;  /* double size if 16 bit      */
   else if (afmt!=AFMT_U8) {
      exit(err_rpt(EINVAL,"Only unsigned 8 and signed 16 bit, supported."));
   }
	
   Ntone=(Tms*samplerate+500)/1000; /* calc. number of samples for Tms */
   if (vflg) {
      fputs((afmt==AFMT_S16_LE)?"16":"8",stdout);
      fputs(" bit samples being generated.\n",stdout);
      fputd(tbuf_size,stdout); 
      fputs(" byte buffer allocated per tone.\n",stdout);
      fputs("Each tone played for ",stdout); fputd(Tms,stdout);
      fputs(" millisecs, ",stdout); fputd(Ntone,stdout);
      fputs(" samples.\n",stdout);
   }

   if (makealltones(argc,argv))
      exit(err_rpt(EINVAL,"Couldn't generatei one of required tone(s)."));
   silence=maketone("OFF",0);     /* create a silence buf for infill at end of play */

/* now either play them or write them to a file......
 */
   
   if (outfile!=NULL) {
      loopflg=0;    /*  cannot loop if writing to file. */
      if (wfmt) {
         st=writewav(fd,NULL,(Nt*Ntone)<<(afmt==AFMT_S16_LE)); 
                         /* total bytes is number of tones times number of */
                         /* samples per tone times 2 if 16bit samples */
      } 
   }
   for (st=i=0; loopflg || (i<Nt); i++) {
      if (i>=Nt) i=i%Nt;
      st|=playtone(fd,tones[i],samplerate,Tms,afmt);
   }
   if (outfile==NULL) {
      DSPend(fd);     
   } else { 
      if (st<0) {
         err_rpt(errno,outfile);
      } else if (vflg) {
         fputd(Nt*Ntone,stdout);
         fputs(" samples written to ",stdout);
	 fputs((wfmt)?"WAVE":"RAW",stdout);
	 fputs(" format file ",stdout); fputs(outfile,stdout);
         fputs("\n",stdout);
      }
   }
   
   for (i=0; i<Nt; i++) { free(tones[i]); }
       
   close(fd);
   exit(0);
}  

/* makealltones(argv,argc)
 * generate the required tones from arguments in the argc arguments in argv
 */

makealltones(argc,argv)
int argc;
char **argv;
{
   char *p,*s,*t;
   int i,n;

   for (Nt=0; Nt<MAXTONES && argc ; Nt++,argc--,*argv++) {
      n=strtol(*argv,&s,10);
      if ((tones[Nt]=maketone(waveform,n))==NULL) return(1);
      for ( ; ispunct(*s); ) {
         p=++s; n=strtol(p,&s,10);
         if ((t=maketone(waveform,n))==NULL) return(1);
         chanmix(tones[Nt],tones[Nt],100,t,100,samplerate,afmt);
         free(t);
      }
   }
   return(0);
}

/* maketone(wf,fr)   generate a tone of frequency fr, waveform wf,
 *   does a malloc to get a new tone buffer.
 */

unsigned char *maketone(wf,fr)
char *wf;
int fr;

{
   unsigned char *bf;
   int n;
   
   if ((bf=(unsigned char *)malloc(tbuf_size))==NULL) return(NULL);
   n=generate((fr)?wf:"OFF",bf,tbuf_size,fr,100,samplerate,-1,afmt);
   return((n)?bf:NULL);
}

/* playtone(fd,bf,N,t,afmt)  play t millisecs of tone in buffer bf containing
 *                 N samples of afmt format, card is set to play at samplerate
 *                 samples per sec.
 */

playtone(fd,bf,N,t,afmt)
int fd;
unsigned char *bf;
int N,t,afmt;
{
   int i,l,m,st,b16;
   unsigned char *p,*q;
   
   b16=(afmt==AFMT_S16_LE);   /* =1 for 16 bit, =0 for 8 bit. */
   m=(t*samplerate+500)/1000; /* calculate the number of samples to play */
   DSPbytecount+=(m<<b16);    /* update total byte count */
   for (st=0; m>=N; m-=N) {
      st|=write(fd,bf,N<<b16);
   }
   if (m) st|=write(fd,bf,m<<b16);
   return(st);
}

DSPend(fd)
int fd;
{
   int n;

   n=frag_size-(DSPbytecount%frag_size);
   if (n) write(fd,silence,n);      /* fill in current fragment buffer */
}

/*
 *  playloop(fd,bf,bfn)   bf is a sample buffer of size bfn bytes.
 *    
 */

playloop(fd,bf,bfn)
int fd;
unsigned char *bf;
int bfn;
{
   int i,bn;
   unsigned char *p;
   
   bn=bfn/10;
   for (;;) {
      for ( p=bf,i=0; i< bfn; i+=bn) { 
	 if (write(fd,p+i,bn) < 0) return(0);
	 /* check here if need to supend output or handle keypresses etc*/
      }
      if ((i!=bfn) && (write(fd,p+i-bn,bfn-i+bn)<0)) return(0);
   }
}

/*
 * DACopen(char *fnm, char *mode, int *samples, int *fmt, int *stereo)
 * 
 *   open dspfile for read "r" or write "w", and set samples per sec
 *    as sampling rate - note we get pointer for samples so we can 
 *    return the actual samplerate set. Stereo is non zero for stereo,
 *    and fmt is the format (8 or 16 bit)
 *    and again these are pointers, so actually settings can be returned.
 *   return file descriptor or -1 on error.
 */

DACopen(fnm,mode,samples,fmt,stereo)
char *fnm;
char *mode;
int *samples, *stereo, *fmt;
{
   int fd;
   int m;
   
   if (*mode=='r') m=O_RDONLY;
   else if (*mode=='w') m=O_WRONLY;
   else {
      errno=EINVAL; return(-1);
   }
   if ((fd = open (fnm,m,0)) >= 0) {  /* params must be set in this order */
      if (ioctl(fd, SNDCTL_DSP_SETFMT, fmt)>=0) { 
	 if (ioctl(fd, SNDCTL_DSP_STEREO, stereo)>=0) { 
	    if (ioctl(fd, SNDCTL_DSP_SPEED, samples)>=0) { 
               if (ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &frag_size)>=0) {
                  DSPbytecount=0;
	          return(fd); 
               }
	    }
	 }
      }
   }
   return(-1);
}

/* writewav(fd,buf,N)  write WAV file header to fd for N samples.
 *    if buf is not NULL then write N samples from buf to fd
 */

writewav(fd,buf,N)
int fd;
char *buf;
int N;

{
   WaveHeader wh;

   wh.main_chunk = RIFF;
   wh.length     = N + sizeof(WaveHeader) - 8; 
   wh.chunk_type = WAVE;
   wh.sub_chunk  = FMT;
   wh.sc_len     = 16;
   wh.format     = PCM_CODE;
   wh.modus      = (stereo) ? 2 : 1;
   wh.sample_fq  = samplerate;
   wh.byte_p_spl = (afmt==AFMT_U8) ? 1 : 2;
   wh.byte_p_sec = samplerate * wh.modus * wh.byte_p_spl;
   wh.bit_p_spl  = (afmt==AFMT_U8) ? 8 : 16;
   wh.data_chunk = DATA;
   wh.data_length= N;
   if (write(fd, &wh, sizeof(WaveHeader))<0) return(-1);
   return((buf==NULL)?0:write(fd, buf, N));
}
   
